from __future__ import print_function,  absolute_import
import underworld as uw
import numpy as np
from mpi4py import MPI
from scipy.interpolate import interp1d, interp2d
from UWGeodynamics._material import Material

comm = MPI.COMM_WORLD
rank = comm.Get_rank()


class LecodeIsostasy(object):
    """LecodeIsostasy"""

    def __init__(self, reference_mat=None, average=False,
                 surface=None, maskedMat=None,
                 vertical_walls_conditions=None):
        """LecodeIsostasy Init Method

        Parameters
        ----------

        reference_mat : reference material at the base of the Model
                        (The Material that enters or leaves the Model
                         to account for isostasy)
        average : (bool) if True, calculate an average velocity at the base of the Model.
        surface : (optional) Initial position of the surface of the material columns. The
                  top of the Model is used by default and should be fine in most cases.
        maskedMat : (optional) Mask a material or a list of materials.
                    The masked material will not interfere in the calculation of the densities.
        vertical_walls_conditions : (optional) Existing boundary conditions on the
                                    lateral walls.

        Returns
        -------

        GEO.LecodeIsostasy object.

        """

        self.mesh = None
        self.swarm = None
        self._mesh_advector = None
        self.velocityField = None
        self._densityFn = None
        self.materialIndexField = None

        if not isinstance(reference_mat, Material):
            raise ValueError("""reference_mat must be a Material, index of
                             Material has been deprecated""")

        self.reference_mat = reference_mat
        self.average = average
        self.surface = surface
        self.maskedMat = list(maskedMat) if maskedMat else list()
        self._maskedMatIndices = [mat.index for mat in self.maskedMat]
        self.initialized = False

        # Make the class aware of the conditions on the vertical walls
        if vertical_walls_conditions:
            if not isinstance(vertical_walls_conditions, dict):
                raise TypeError(
                    """vertical_walls_conditions must be of type
                       'dict'""")
            options = ["left", "right", "front", "back"]
            for key in vertical_walls_conditions.keys():
                if key not in options:
                    raise KeyError("""
                                   Wrong dictionary key, options are:
                                   {0}""".format(options))
            self.vertical_walls_conditions = vertical_walls_conditions

    def solve(self):

        if not self.initialized:
            self._check_all_defined()

        if self.mesh.dim == 2:
            self._lecode_tools_isostasy2D()

        if self.mesh.dim == 3:
            self._lecode_tools_isostasy3D()

    def _check_all_defined(self):
        if not self.mesh:
            raise ValueError("Please link a Mesh to the Isostasy solver")
        if not self.swarm:
            raise ValueError("Please link a Swarm to the Isostasy solver")
        if not self.velocityField:
            raise ValueError("""
                Please link a Velocity Field to the Isostasy solver""")
        if not self.materialIndexField:
            raise ValueError("""
                Please link a Material Field to the Isostasy solver""")
        if not self.materialIndexField:
            raise ValueError("""
                Please link a Material Field to the Isostasy solver""")
        if not self._densityFn:
            raise ValueError("""
                Please link a Density Field to the Isostasy solver""")
        if not self.reference_mat:
            raise ValueError("""
                Please define a reference Material for the Isostasy solver""")

        # Create utilities
        self.MaterialIndexFieldFloat = self.swarm.add_variable(
            dataType="double", count=1)
        self.DensityVar = uw.mesh.MeshVariable(self.mesh, nodeDofCount=1)
        self.MaterialVar = uw.mesh.MeshVariable(self.mesh, nodeDofCount=1)
        self.projectorDensity = uw.utils.MeshVariable_Projection(
            self.DensityVar,
            self._densityFn,
            type=0)
        self.projectorMaterial = uw.utils.MeshVariable_Projection(
            self.MaterialVar, self.MaterialIndexFieldFloat, type=0)

        if not self.mesh._cself.isRegular:
            if self._mesh_advector:
                if self._mesh_advector.axis > 0:
                    raise TypeError("""You are using an irregular mesh: \
                                    isostasy module only works with regular
                                    meshes""")
            else:
                raise TypeError("""You are using an irregular mesh: \
                                isostasy module only works with regular
                                meshes""")

        if (self.surface is not None and
           not isinstance(self.surface, uw.swarm._swarm.Swarm)):
            raise TypeError("'surface' must be a tuple'")

        self.initialized = True

    def _lecode_tools_isostasy2D(self):

        self.MaterialIndexFieldFloat.data[...] = (
            np.rint(self.materialIndexField.data.astype("float"))
        )

        self.projectorDensity.solve()
        self.projectorMaterial.solve()

        sep_velocities_nodes, _ = self._get_sep_velocities2D()
        botMeanDensities, botMeanDensities0 = self._get_average_densities2D()

        basal_velocities = -1.0 * botMeanDensities * sep_velocities_nodes / botMeanDensities0

        if self.average:
            basal_velocities = np.ones((basal_velocities.shape)) * np.mean(basal_velocities)

        base = self.mesh.specialSets["MinJ_VertexSet"]

        if base:

            # Look at the vertical conditions on the vertical walls
            # if one has been defined, remove the corners from the base
            # and do not change the velocity on those nodes.
            if self.vertical_walls_conditions:
                if self.vertical_walls_conditions["left"]:
                    if self.vertical_walls_conditions["left"][-1] is not None:
                        left = self.mesh.specialSets["MinI_VertexSet"]
                        if left:
                            base -= left
                if self.vertical_walls_conditions["right"]:
                    if self.vertical_walls_conditions["right"][-1] is not None:
                        right = self.mesh.specialSets["MaxI_VertexSet"]
                        if right:
                            base -= right

            bot_ids = base.data[base.data < self.mesh.nodesLocal]
            node_gids = self.mesh.data_nodegId[bot_ids].flatten()
            self.velocityField.data[bot_ids, 1] = basal_velocities[node_gids]

        self.velocityField.syncronise()

    def _lecode_tools_isostasy3D(self):

        self.MaterialIndexFieldFloat.data[...] = (
            np.rint(self.materialIndexField.data.astype("float"))
        )

        self.projectorDensity.solve()
        self.projectorMaterial.solve()

        sep_velocities_nodes, _ = self._get_sep_velocities3D()
        botMeanDensities, botMeanDensities0 = self._get_average_densities3D()

        basal_velocities = -1.0 * botMeanDensities * sep_velocities_nodes / botMeanDensities0

        if self.average:
            basal_velocities = np.ones((basal_velocities.shape)) * np.mean(basal_velocities)

        base = self.mesh.specialSets["MinK_VertexSet"]
        if base:

            # Look at the vertical conditions on the vertical walls
            # if one has been defined, remove the corners from the base
            # and do not change the velocity on those nodes.
            if self.vertical_walls_conditions:
                if self.vertical_walls_conditions["left"]:
                    if self.vertical_walls_conditions["left"][-1] is not None:
                        left = self.mesh.specialSets["MinI_VertexSet"]
                        if left:
                            base -= left
                if self.vertical_walls_conditions["right"]:
                    if self.vertical_walls_conditions["right"][-1] is not None:
                        right = self.mesh.specialSets["MaxI_VertexSet"]
                        if right:
                            base -= right
                if self.vertical_walls_conditions["front"]:
                    if self.vertical_walls_conditions["front"][-1] is not None:
                        front = self.mesh.specialSets["MinJ_VertexSet"]
                        if front:
                            base -= front
                if self.vertical_walls_conditions["back"]:
                    if self.vertical_walls_conditions["back"][-1] is not None:
                        back = self.mesh.specialSets["MaxJ_VertexSet"]
                        if back:
                            base -= back
            bot_ids = base.data[base.data < self.mesh.nodesLocal]
            node_gids = self.mesh.data_nodegId[bot_ids].flatten()
            self.velocityField.data[bot_ids, -1] = basal_velocities.flatten()[node_gids]

        self.velocityField.syncronise()

    def _get_sep_velocities2D(self):

        ncol, nrow = self.mesh.elementRes
        if self.mesh.elementType == "Q2":
            fact = 2
        elif self.mesh.elementType == "Q1":
            fact = 1
        ncol *= fact
        nrow *= fact

        # Create some work arrays.
        local_top_vy = np.zeros((ncol + 1,))
        local_bot_vy = np.zeros((ncol + 1,))
        global_top_vy = np.zeros((ncol + 1,))
        global_bot_vy = np.zeros((ncol + 1,))

        local_heights = np.zeros((ncol + 1,))
        global_heights = np.zeros((ncol + 1,))

        # Get the ids of the nodes belonging to the top and bottom of the
        # global domain.
        top_ids = self.mesh.specialSets["MaxJ_VertexSet"]
        bot_ids = self.mesh.specialSets["MinJ_VertexSet"]

        if self.surface is not None:

            surfLocals, surfGlobals = self._get_surface_nodes()
            if surfLocals.size > 0:
                # Get an I,J representation of the node coordinates
                Jpositions = (surfGlobals.astype("float") / (ncol + 1)).astype("int")
                Ipositions = surfGlobals - Jpositions * (ncol + 1)

                # Load the top-velocities in the local array with global dimensions.
                local_top_vy[Ipositions.flatten()] = self.velocityField.data[surfLocals, 1]
                local_heights[Ipositions.flatten()] += self.mesh.data[surfLocals][:,1]

        # If the local domain contains some of the top_ids, proceed:
        elif top_ids:

            # We must get rid of the shadow nodes
            top_ids = top_ids.data[top_ids.data < self.mesh.nodesLocal]

            # Get the cooresponding global ids
            node_gids = self.mesh.data_nodegId[top_ids]

            # Get y-velocities at the top
            topVelocities_nodes = self.velocityField.data[top_ids,1]
            heights_nodes = self.mesh.data[top_ids,1]

            # Get an I,J representation of the node coordinates
            Jpositions = (node_gids.astype("float") / (ncol + 1)).astype("int")
            Ipositions = node_gids - Jpositions * (ncol + 1)

            # Load the top-velocities in the local array with global dimensions.
            j = 0
            for pos in Ipositions:
                local_top_vy[pos] = topVelocities_nodes[j]
                local_heights[pos] += heights_nodes[j]
                j+=1

        # If the local domain contains some of the bot_ids, proceed:
        if bot_ids:

            # We must get rid of the shadow nodes
            bot_ids = bot_ids.data[bot_ids.data < self.mesh.nodesLocal]

            # Get the cooresponding global ids
            node_gids = self.mesh.data_nodegId[bot_ids]

            # Get y-velocities at the bottom
            botVelocities_nodes = self.velocityField.data[bot_ids,1]
            heights_nodes = self.mesh.data[bot_ids,1]

            # Get an I,J representation of the node coordinates
            Jpositions = (node_gids.astype("float") / (ncol + 1)).astype("int")
            Ipositions = node_gids - Jpositions * (ncol + 1)

            # Load the bottom-velocities in the local array with global dimensions.
            j = 0
            for pos in Ipositions:
                local_bot_vy[pos] = botVelocities_nodes[j]
                local_heights[pos] -= heights_nodes[j]
                j+=1

        # reduce local arrays into global_array
        comm.Allreduce(local_top_vy, global_top_vy)
        comm.Allreduce(local_bot_vy, global_bot_vy)
        comm.Allreduce(local_heights, global_heights)
        comm.Barrier()

        # 3-nodes mean average
        global_top_vy = (np.roll(global_top_vy, -1) + global_top_vy + np.roll(global_top_vy, 1)) / 3.0
        global_bot_vy = (np.roll(global_bot_vy, -1) + global_bot_vy + np.roll(global_bot_vy, 1)) / 3.0
        global_heights = (np.roll(global_heights, -1) + global_heights + np.roll(global_heights, 1)) / 3.0

        # Calculate and return sep velocities
        return global_top_vy - global_bot_vy, global_heights

    def _get_sep_velocities3D(self):

        nx, ny, nz = self.mesh.elementRes
        if self.mesh.elementType == "Q2":
            fact = 2
        elif self.mesh.elementType == "Q1":
            fact = 1
        nx *= fact
        ny *= fact
        nz *= fact
        GlobalIndices3d = np.indices((nz + 1, ny + 1, nx + 1))

        # Create some work arrays.
        local_top_vy = np.zeros(((ny + 1), (nx + 1)))
        local_bot_vy = np.zeros(((ny + 1), (nx + 1)))
        global_top_vy = np.zeros(((ny + 1), (nx + 1)))
        global_bot_vy = np.zeros(((ny + 1), (nx + 1)))

        local_heights = np.zeros(((ny + 1), (nx + 1)))
        global_heights = np.zeros(((ny + 1), (nx + 1)))

        # Get the ids of the nodes belonging to the top and bottom of the
        # global domain.
        top_ids = self.mesh.specialSets["MaxK_VertexSet"]
        bot_ids = self.mesh.specialSets["MinK_VertexSet"]

        if self.surface is not None:

            _, surfaceNodes = self._get_surface_nodes()
            if surfaceNodes.size > 0:
                ix = np.in1d(np.arange(self.mesh.nodesGlobal), surfaceNodes)
                Ipositions = GlobalIndices3d[2].flatten()[ix]
                Jpositions = GlobalIndices3d[1].flatten()[ix]

                # Load the top-velocities in the local array with global dimensions.
                k = 0
                for i, j in zip(Ipositions, Jpositions):
                    local_top_vy[j, i] = self.velocityField.data[surfaceNodes][k]
                    local_heights[j, i] += self.mesh.data[surfaceNodes,2][k]
                    k += 1

        # If the local domain contains some of the top_ids, proceed:
        elif top_ids:

            # We must get rid of the shadow nodes
            top_ids = top_ids.data[top_ids.data < self.mesh.nodesLocal]

            # Get the cooresponding global ids
            node_gids = self.mesh.data_nodegId[top_ids]

            # Get y-velocities at the top
            topVelocities_nodes = self.velocityField.data[top_ids, -1]
            heights_nodes = self.mesh.data[top_ids, -1]

            # Get an I,J representation of the node coordinates

            # FROM Numpy 1.13> ... more efficient ?
            #ix = np.isin(GlobalIndices3d, node_gids)
            #Kpositions, Jposition, Iposition = np.where(ix)
            ix = np.in1d(np.arange(self.mesh.nodesGlobal), node_gids)
            Ipositions = GlobalIndices3d[2].flatten()[ix]
            Jpositions = GlobalIndices3d[1].flatten()[ix]

            # Load the top-velocities in the local array with global dimensions.
            k = 0
            for i, j in zip(Ipositions, Jpositions):
                local_top_vy[j, i] = topVelocities_nodes[k]
                local_heights[j, i] += heights_nodes[k]
                k += 1

        # If the local domain contains some of the bot_ids, proceed:
        if bot_ids:

            # We must get rid of the shadow nodes
            bot_ids = bot_ids.data[bot_ids.data < self.mesh.nodesLocal]

            # Get the cooresponding global ids
            node_gids = self.mesh.data_nodegId[bot_ids]

            # Get y-velocities at the bottom
            botVelocities_nodes = self.velocityField.data[bot_ids,-1]
            heights_nodes = self.mesh.data[bot_ids,-1]

            # Get an I,J representation of the node coordinates
            ix = np.in1d(np.arange(self.mesh.nodesGlobal), node_gids)
            Ipositions = GlobalIndices3d[2].flatten()[ix]
            Jpositions = GlobalIndices3d[1].flatten()[ix]

            # Load the top-velocities in the local array with global dimensions.
            k = 0
            for i, j in zip(Ipositions, Jpositions):
                local_bot_vy[j, i] = botVelocities_nodes[k]
                local_heights[j, i] -= heights_nodes[k]
                k += 1

        # reduce local arrays into global_array
        comm.Allreduce(local_top_vy, global_top_vy)
        comm.Allreduce(local_bot_vy, global_bot_vy)
        comm.Allreduce(local_heights, global_heights)
        comm.Barrier()

        # Calculate and return sep velocities
        return global_top_vy - global_bot_vy, global_heights

    def _get_surface_nodes(self):
        """ Get the closest mesh nodes to a surface
            The function returns the global ids by default
        """

        procCount = comm.allgather(self.surface.particleLocalCount)
        particleGlobalCount = np.sum(procCount)

        offset = 0
        for i in range(rank):
            offset += procCount[i]

        if self.mesh.dim == 2:

            localSurface = np.zeros((particleGlobalCount, 2), self.surface.particleCoordinates.data.dtype)
            globalSurface = np.zeros((particleGlobalCount, 2), self.surface.particleCoordinates.data.dtype)

            if self.surface.particleLocalCount > 0:
                localSurface[offset:offset + self.surface.particleLocalCount] = self.surface.particleCoordinates.data[:]

            comm.Allreduce(localSurface, globalSurface)
            comm.Barrier()

            surface = globalSurface
            surface.sort(axis=0)

            # Build linearly interpolated function from surface points
            f = interp1d(surface[:, 0], surface[:, 1], bounds_error=False, fill_value="extrapolate")
            # Get y positions on the surface
            ypos = f(self.mesh.data[:, 0])
            dy = np.abs((self.mesh.maxCoord[1] - self.mesh.minCoord[1]) / self.mesh.elementRes[1])
            localIds = np.arange(self.mesh.nodesDomain)[np.abs(self.mesh.data[:, 1] - ypos) < (dy / 2.0)]
            globalIds = self.mesh.data_nodegId[localIds]
            return localIds, globalIds

        if self.mesh.dim == 3:

            localSurface = np.zeros((particleGlobalCount, 3), self.surface.particleCoordinates.data.dtype)
            globalSurface = np.zeros((particleGlobalCount, 3), self.surface.particleCoordinates.data.dtype)

            if surface.particleLocalCount > 0:
                localSurface[offset:offset + surface.particleLocalCount] = self.surface.particleCoordinates.data[:]

            comm.Allreduce(localSurface, globalSurface)
            comm.Barrier()

            surface = globalSurface
            # Sort the array
            surface = surface[np.lexsort(np.transpose(surface)[::-1])]

            f = interp2d(surface[:, 0], surface[:, 1], surface[:, 2])
            # Get y positions on the surface
            zpos = f(self.mesh.data[:, 0], self.mesh.data[:, 1])
            dz = np.abs((self.mesh.maxCoord[2] - self.mesh.minCoord[2]) / self.mesh.elementRes[2])
            localIds = np.arange(self.mesh.nodesLocal)[np.abs(self.mesh.data[:, 2] - zpos) < dz / 2.0]
            globalIds = self.mesh.data_nodegId[localIds]
            return localIds, globalIds

    def _get_average_densities2D(self):

        ncol, nrow = self.mesh.elementRes
        if self.mesh.elementType == "Q2":
            fact = 2
        elif self.mesh.elementType == "Q1":
            fact = 1
        ncol *= fact
        nrow *= fact

        # Create some work arrays.
        global_densities = np.zeros((nrow + 1) * (ncol + 1))
        local_densities = np.zeros((nrow + 1) * (ncol + 1))
        local_materials = np.zeros((nrow + 1) * (ncol + 1))
        global_materials = np.zeros((nrow + 1) * (ncol + 1))

        # Load the densities and material into the local_densities arrays
        # data_nodegId as domain size (local+node) so we need to only take
        # the local nodes otherwise the reduce operation will results in higher
        # values where domains overlaps...
        local_densities[self.mesh.data_nodegId[:self.mesh.nodesLocal]] = self.DensityVar.data[:self.mesh.nodesLocal]
        local_materials[self.mesh.data_nodegId[:self.mesh.nodesLocal]] = self.MaterialVar.data[:self.mesh.nodesLocal]

        # Reduce local_densities arrays to global_densities
        comm.Allreduce(local_densities, global_densities)
        comm.Allreduce(local_materials, global_materials)
        # It seems that the MPI implementation to not systematically impose a
        # barrier inside a reduce operation. That is weird but just to be safe:
        comm.Barrier()

        # Reshape the arrays
        global_densities = global_densities.reshape(((nrow + 1), (ncol + 1)))
        global_materials = global_materials.reshape(((nrow + 1), (ncol + 1)))
        # Convert material values to closest integers.
        global_materials = global_materials.astype("int")

        # Calculate Mean densities at the bottom
        if self.maskedMat:
            mask = np.in1d(global_materials, self.maskedMatIndices)
            maskedArray = np.ma.masked_array(global_densities, mask)
            botMeanDensities = maskedArray.mean(axis=0)
        else:
            botMeanDensities = np.mean(global_densities, 0)

        # Calculate Mean densities at the bottom (reference_mat only)
        botMeanDensities0 = np.ma.array(global_densities, mask=(global_materials != self.reference_mat.index))
        if botMeanDensities0.count() == 0:
            raise ValueError("""I am trying hard here but it looks like you
                             don't have any material with index {0} in your model.
                             Please check your set up. The resolution might
                             also be insufficient to resolve the
                             material...""".format(self.reference_mat.name))

        botMeanDensities0 = botMeanDensities0.mean(axis=0)
        botMeanDensities0 = np.array(botMeanDensities0)

        botMeanDensities = (np.roll(botMeanDensities, -1) + botMeanDensities + np.roll(botMeanDensities, 1)) / 3.0
        botMeanDensities0 = (np.roll(botMeanDensities0, -1) + botMeanDensities0 + np.roll(botMeanDensities0, 1)) / 3.0

        # return bottom Densities and Densities0
        return botMeanDensities, botMeanDensities0

    def _get_average_densities3D(self):

        nx, ny, nz = self.mesh.elementRes
        if self.mesh.elementType == "Q2":
            fact = 2
        elif self.mesh.elementType == "Q1":
            fact = 1
        nx *= fact
        ny *= fact
        nz *= fact

        # Create some work arrays.
        global_densities = np.zeros((nx+1)*(ny+1)*(nz+1))
        local_densities = np.zeros((nx+1)*(ny+1)*(nz+1))
        local_materials = np.zeros((nx+1)*(ny+1)*(nz+1))
        global_materials = np.zeros((nx+1)*(ny+1)*(nz+1))

        # Load the densities and material into the local_densities arrays
        local_densities[self.mesh.data_nodegId[:self.mesh.nodesLocal]] = self.DensityVar.data[:self.mesh.nodesLocal]
        local_materials[self.mesh.data_nodegId[:self.mesh.nodesLocal]] = self.MaterialVar.data[:self.mesh.nodesLocal]

        # Reduce local_densities arrays to global_densities
        comm.Allreduce(local_densities, global_densities)
        comm.Allreduce(local_materials, global_materials)
        comm.Barrier()

        # Reshape the arrays
        global_densities = global_densities.reshape(((nz + 1), (ny + 1), (nx + 1)))
        global_materials = global_materials.reshape(((nz + 1), (ny + 1), (nx + 1)))
        # Convert material values to closest integers.
        global_materials = np.rint(global_materials).astype("int")

        # Calculate Mean densities at the bottom
        if self.maskedMat:
            mask = np.in1d(global_materials, self.maskedMatIndices)
            maskedArray = np.ma.masked_array(global_densities, mask)
            botMeanDensities = maskedArray.mean(axis=0)
        else:
            botMeanDensities = np.mean(global_densities, 0)

        # Calculate Mean densities at the bottom (reference_mat only)
        botMeanDensities0 = np.ma.array(global_densities, mask=(global_materials != self.reference_mat.index))
        if botMeanDensities0.count() == 0:
            raise ValueError("""I am trying hard here but it looks like you
                             don't have any material with index {0} in your model.
                             Please check your set up. The resolution might
                             also be insufficient to resolve the
                             material...""".format(self.reference_mat.name))
        botMeanDensities0 = botMeanDensities0.mean(axis=0)
        botMeanDensities0 = np.array(botMeanDensities0)

        # return bottom Densities and Densities0
        return botMeanDensities, botMeanDensities0




from __future__ import print_function,  absolute_import
from ._utils import TransformedDict
from ._utils import u

COEFFICIENTS = TransformedDict()
COEFFICIENTS["[length]"] = 1.0 * u.meter
COEFFICIENTS["[mass]"] = 1.0 * u.kilogram 
COEFFICIENTS["[time]"] = 1.0 * u.year
COEFFICIENTS["[temperature]"] = 1.0 * u.degK
COEFFICIENTS["[substance]"] = 1.0 * u.mole



from __future__ import print_function,  absolute_import
import numpy as np
import underworld as uw
import underworld.function as fn
from .scaling import nonDimensionalize as nd


class Shape(object):
    """Base Class for Shape objects"""

    def __init__(self):
        self._fn = None
        return

    @property
    def fn(self):
        return self._fn

    @fn.setter
    def fn(self, value):
        self._fn = value

    def __and__(self, B):
        newShape = Shape()
        newShape.fn = self._fn & B._fn
        return newShape

    def __add__(self, B):
        newShape = CombinedShape([self, B])
        return newShape

    def __or__(self, B):
        newShape = Shape()
        newShape.fn = self._fn | B._fn
        return newShape


class Polygon(Shape):
    """Polygon Shape Class"""

    def __init__(self, vertices):
        """Create a polygon shape

        Parameters
        ----------

        vertices : vertices of the polygon as (x,y) pairs.

        Returns
        -------

        Polygon Shape Class
        """
        self.vertices = vertices
        self.top = min([y for x, y in self.vertices])
        self.bottom = max([y for x, y in self.vertices])
        vertices = [(nd(x), nd(y)) for x, y in self.vertices]
        self._fn = uw.function.shape.Polygon(np.array(vertices))


class HalfSpace(Shape):
    """ Class to define a HalfSpace

        A plan defined by a normal vector is used to split the space
        in two half spaces. By default the origin of the coordinate
        system is at (0., 0.).

        Particles tested against this class are assigned a boolean value.
    """

    def __init__(self, normal, origin=None, reverse=False):
        """ HalfSpace

        Parameters:
        -----------

        normal: A vector defining the normal to the plan.
        origin: Origin
        reverse: by default, particles tested against this class are
                 assigned "True" if they lay on or below the plan.
                 You can reverse than behavior by setting reverse=True.

        Returns:
        --------

        A UWGeodynamics Shape object.

        """

        if isinstance(normal, (tuple, list)):
            self.normal = fn.misc.constant([float(nd(val)) for val in normal])
        else:
            raise ValueError("{0} must be a list or tuple".format(normal))

        if isinstance(origin, (tuple, list)):
            self.origin = fn.misc.constant([float(nd(val)) for val in origin])
        else:
            self.origin = fn.misc.constant([0.] * len(normal))

        self.reverse = reverse

    @property
    def _fn(self):
        coords = fn.input()
        new_coords = coords - self.origin
        func = fn.math.dot(self.normal, new_coords)

        # True if below, False if above
        if not self.reverse:
            conditions = [(func <= 0., True), (func > 0., False)]
        else:
            conditions = [(func >= 0., True), (func < 0., False)]

        return fn.branching.conditional(conditions)


class CombinedShape(Shape):
    """CombinedShape class"""

    def __init__(self, shapes):
        """Combine a list of shapes into one

        Parameters
        ----------

        shapes : list of Shape objects

        Returns
        -------

        An UWGeodynamics Shape object
        """
        self.shapes = shapes
        self.top = max([shape.top for shape in shapes])
        self.bottom = min([shape.top for shape in shapes])

    @property
    def _fn(self):
        import operator
        import functools
        self._fnlist = []
        for shape in self.shapes:
            self._fnlist.append(shape._fn)
        func = functools.reduce(
            operator.or_,
            self._fnlist,
            fn.misc.constant(False))
        return func


class MultiShape(CombinedShape):

    def __init__(self, shapes):

        Warning("MultiShape is now deprecated, use CombinedShape instead")
        super(MultiShape, self).__init__(shapes)


class IntersectShape(Shape):
    """IntersectShape class"""

    def __init__(self, shapes):
        """ Take the intersection of some shapes

        Parameters
        ----------

        shapes : A list of Shapes

        Returns
        -------

        An UWGeodynamics Shape object
        """
        self.shapes = shapes

    @property
    def _fn(self):
        import operator
        import functools
        self._fnlist = []
        for shape in self.shapes:
            self._fnlist.append(shape._fn)
        func = functools.reduce(
            operator.and_,
            self._fnlist,
            fn.misc.constant(True))
        return func


class Layer(Shape):
    """Layer 2D"""

    def __init__(self, top, bottom):
        """Create a 2D Layer object

        Parameters
        ----------

        top : top of the layer
        bottom : bottom of the layer

        Returns
        -------

        AN UWGeodynamics Shape object
        """
        self.top = top
        self.bottom = bottom

    @property
    def _fn(self):
        coord = fn.input()
        func = ((coord[1] <= nd(self.top)) &
                (coord[1] >= nd(self.bottom)))
        return func


class Layer3D(Layer):
    """Layer3D"""

    def __init__(self, top, bottom):
        """Create a 3D Layer object

        Parameters
        ----------

        top : top of the layer
        bottom : bottom of the layer

        Returns
        -------

        AN UWGeodynamics Shape object
        """
        self.top = top
        self.bottom = bottom

    @property
    def _fn(self):
        coord = fn.input()
        func = ((coord[2] <= nd(self.top)) &
                (coord[2] >= nd(self.bottom)))
        return func

class Layer2D(Layer):

    def __init__(self, top, bottom):

        Warning("Layer2D is now deprecated, use Layer instead")
        super(Layer2D, self).__init__(top, bottom)

class Box(Shape):
    """Box"""

    def __init__(self, top, bottom, minX=0., maxX=0., minY=None, maxY=None):
        """Create a Box Shape

        Parameters
        ----------

        top : Top of the Box
        bottom : Bottom of the Box
        minX : Minimum extent of the Box along the x-axis
        maxX : Maximum extent of the Box along the x-axis

        Only in 3D:

        minY : Minimum extent of the Box along the y-axis
        maxY : Maximum extent of the Box along the y-axis

        Returns
        -------
        """
        self.top = top
        self.bottom = bottom
        self.minX = minX
        self.maxX = maxX
        self.minY = minY
        self.maxY = maxY

    @property
    def _fn(self):
        coord = fn.input()
        if (self.minY is not None) and (self.maxY is not None):
            func = ((coord[1] <= nd(self.maxY)) &
                    (coord[1] >= nd(self.minY)) &
                    (coord[0] <= nd(self.maxX)) &
                    (coord[0] >= nd(self.minX)) &
                    (coord[2] <= nd(self.top)) &
                    (coord[2] >= nd(self.bottom)))
        else:
            func = ((coord[1] <= nd(self.top)) &
                    (coord[1] >= nd(self.bottom)) &
                    (coord[0] <= nd(self.maxX)) &
                    (coord[0] >= nd(self.minX)))
        return func


class Disk(Shape):
    """Disk"""

    def __init__(self, center, radius):
        """Create a Disk shape

        Parameters
        ----------

        center : center of the disk
        radius : radius of the disk

        Returns
        -------

        An UWGeodynamics Shape
        """
        self.center = center
        self.radius = radius
        self.top = center[1] + self.radius
        self.bottom = center[1] - self.radius

    @property
    def _fn(self):
        center = tuple(nd(x) for x in list(self.center))
        radius = nd(self.radius)
        coord = fn.input() - center
        return fn.math.dot(coord, coord) < radius**2


Sphere = Disk


class Annulus(Shape):
    """Annulus"""

    def __init__(self, center, r1, r2):
        """Create an Annulus shape

        Parameters
        ----------

        center : center of the annulus
        r1 : Internal radius
        r2 : External radius

        Returns
        -------

        An UWGeodynamics Shape object
        """
        self.center = center
        self.r1 = r1
        self.r2 = r2
        self.bottom = center[1] - self.r2
        self.top = center[1] + self.r2

    @property
    def _fn(self):
        center = tuple(nd(x) for x in list(self.center))
        r1 = nd(self.r1)
        r2 = nd(self.r2)
        coord = fn.input() - center
        return (fn.math.dot(coord, coord) < r2**2) & (fn.math.dot(coord, coord) > r1**2)



from __future__ import print_function,  absolute_import
import numpy as np
import underworld as uw
import underworld.function as fn
import shapefile
import h5py
import os
import operator as op
from .scaling import nonDimensionalize as nd
from .scaling import Dimensionalize
from .scaling import UnitRegistry as u
from .Underworld_extended import Swarm
from scipy import spatial
from mpi4py import MPI

class PhaseChange(object):

    def __init__(self, condition, result):
        self.condition = condition
        self.result = result

    def fn(self):
        conditions = [(self.condition, 1), (True, 0)]
        return fn.branching.conditional(conditions)


class WaterFill(PhaseChange):

    def __init__(self, sealevel, water_material=None):

        self.condition = fn.input()[1] < nd(sealevel)
        self.result = water_material.index


class PressureSmoother(object):

    def __init__(self, mesh, pressureField):

        self.mesh = mesh
        self.pressureField = pressureField

        self.NodePressure = uw.mesh.MeshVariable(self.mesh, nodeDofCount=1)

        self.Cell2Nodes = uw.utils.MeshVariable_Projection(
            self.NodePressure, self.pressureField, type=0)
        self.Nodes2Cell = uw.utils.MeshVariable_Projection(
            self.pressureField, self.NodePressure, type=0)

    def smooth(self):

        self.Cell2Nodes.solve()
        self.Nodes2Cell.solve()


class PassiveTracers(object):

    def __init__(self, mesh, velocityField, name=None, vertices=None,
                 particleEscape=True, shapeType="line"):

        self.name = name
        self.particleEscape = particleEscape

        for dim, _ in enumerate(vertices):
            vertices[dim] = nd(vertices[dim])

        sizes = np.array([np.array(x).size for x in vertices])
        points = np.zeros((sizes.max(), len(vertices)))

        for dim, _ in enumerate(vertices):
            points[:, dim] = vertices[dim]

        self.swarm = Swarm(mesh=mesh, particleEscape=particleEscape)
        self.swarm.add_particles_with_coordinates(points)

        self.advector = uw.systems.SwarmAdvector(swarm=self.swarm,
                                                 velocityField=velocityField,
                                                 order=2)
        indices = np.arange(self.swarm.particleLocalCount)
        rank = uw.rank()
        ranks = np.repeat(rank, self.swarm.particleLocalCount)
        pairs = np.array(list(zip(ranks, indices)), dtype=[("a", np.int32),
                                                           ("b", np.int32)])
        # Get rank
        self.global_index = self.swarm.add_variable(dataType="long", count=1)
        self.global_index.data[:, 0] = pairs.view(np.int64)

        self.tracked_field = list()

    def integrate(self, dt, **kwargs):
        """ Integrate swarm velocity in time """
        self.advector.integrate(dt, **kwargs)

    def add_tracked_field(self, value, name, units, dataType, count=1,
                          overwrite=True):
        """ Add a field to be tracked """
        if not isinstance(value, fn.Function):
            raise ValueError("%s is not an Underworld function")

        # Check that the tracer does not exist already
        for field in self.tracked_field:
            if (name == field["name"]) or (value == field["value"]):
                if not overwrite:
                    raise ValueError(""" %s name already exist or already tracked
                                     with a different name """ % name)
                else:
                    field["name"] = name
                    field["units"] = units
                    field["value"] = value
                    field["dataType"] = dataType
                    setattr(self, name, self.swarm.add_variable(dataType, count=count))
                    return

        self.tracked_field.append({"value":value,
                                   "name": name,
                                   "units": units,
                                   "dataType": dataType})
        setattr(self, name, self.swarm.add_variable(dataType, count=count))

    def write_to_shapefile(self, filename, units=None, overwrite=False):

        if os.path.exists(filename) and not overwrite:
            r = shapefile.Reader(filename)
            w = shapefile.Writer(r.shapeType)
            # Copy over the existing dbf fields
            w.fields = list(r.fields)
            # Copy over the existing dbf records
            w.records.extend(r.records())
            # Copy over the existing polygons
            w._shapes.extend(r.shapes())

        else:

            w = shapefile.Writer(shapeType=shapefile.POLYLINEZ)
            w.field("name", "C")
            w.field("units", "C")

        fact = 1.0
        if units:
            fact = Dimensionalize(1.0, units=units)
            fact = fact.magnitude

        x = self.swarm.particleCoordinates.data[:, 0] * fact
        y = self.swarm.particleCoordinates.data[:, 1] * fact
        w.poly(parts=list(zip(x,y)))
        w.record(self.name, str(units))
        w.save(filename)

    def save(self, outputDir, checkpointID, time):
        """ Save to h5 and create an xdmf file for each tracked field """

        # Save the swarm
        swarm_fname = self.name + '-%s.h5' % checkpointID
        swarm_fpath = os.path.join(outputDir, swarm_fname)

        sH = self.swarm.save(swarm_fpath, units=u.kilometers)

        if uw.rank() == 0:
            filename = self.name + '-%s.xdmf' % checkpointID
            filename = os.path.join(outputDir, filename)

            # First write the XDMF header
            string = uw.utils._xdmfheader()
            string += uw.utils._swarmspacetimeschema(sH, swarm_fname, time)

        uw.barrier()

        # Save global index
        file_prefix = os.path.join(
            outputDir, self.name + '_global_index-%s' % checkpointID)
        handle = self.global_index.save('%s.h5' % file_prefix)

        if uw.rank() == 0:
            string += uw.utils._swarmvarschema(handle, "global_index")
        uw.barrier()

        # Save each tracked field
        for field in self.tracked_field:

            file_prefix = os.path.join(
                outputDir,
                self.name + "_" + field["name"] + '-%s' % checkpointID)

            obj = getattr(self, field["name"])
            obj.data[...] = field["value"].evaluate(self.swarm)
            handle = obj.save('%s.h5' % file_prefix, units=field["units"])

            if uw.rank() == 0:
                # Add attribute to xdmf file
                string += uw.utils._swarmvarschema(handle, field["name"])

        uw.barrier()

        # get swarm parameters - serially read from hdf5 file to get size

        if uw.rank() == 0:
            with h5py.File(name=swarm_fpath, mode="r") as h5f:
                dset = h5f.get('data')
                if dset is None:
                    raise RuntimeError("Can't find 'data' in file '{}'.\n".format(swarm_fname))
                globalCount = len(dset)
                dim = self.swarm.mesh.dim

            string += "\t<Attribute Type=\"Scalar\" Center=\"Node\" Name=\"Coordinates\">\n"
            string += """\t\t\t<DataItem Format=\"HDF\" NumberType=\" Float\"
                         Precision=\"8\" Dimensions=\"{0} {1}\">{2}:/data</DataItem>\n""".format(globalCount, dim, swarm_fname)
            string += "\t</Attribute>\n"

            # Write the footer to the xmf
            string += uw.utils._xdmffooter()

            # Write the string to file - only proc 0
            xdmfFH = open(filename, "w")
            xdmfFH.write(string)
            xdmfFH.close()
        uw.barrier()


class Balanced_InflowOutflow(object):

    def __init__(self, vtop, top, pt1, pt2, ynodes=None,
                 tol=1e-12, nitmax=200,
                 nitmin=3, default_vel=0.0):
        """ Calculate Bottom velocity as for Huismans et al. velocity boundary
            conditions such as the total volume is conserved.
            Return the velocity profile.

            NOTE. The current version implies uniform dy.

            Input:

            Vtop     Top Velocity condition
            pt1, pt2 Top and Bottom location of the transition zone where the
                     velocity linearly decreases from Vtop to VBottom.
            ynodes   Coordinates of the nodes in the y-direction.

            Output:

            velocities numpy array.

            """

        self.vtop = vtop
        self.top = top
        self.pt1 = pt1
        self.pt2 = pt2
        self.ynodes = ynodes
        self.tol = tol
        self.nitmax = nitmax
        self.nitmin = nitmin
        self.default_vel = default_vel


    def _get_side_flow(self):

        Vtop = nd(self.vtop)
        top = nd(self.top)
        pt1 = nd(self.pt1)
        pt2 = nd(self.pt2)
        y = nd(self.ynodes)
        tol = self.tol
        nitmin = self.nitmin
        default_vel = nd(self.default_vel)

        # locate index of closest node coordinates
        top_idx = np.argmin((y - top)**2)
        pt1_idx = np.argmin((y - pt1)**2)
        pt2_idx = np.argmin((y - pt2)**2)

        # do some initialization
        velocity = np.ones(y.shape) * Vtop
        Vmin = -Vtop
        Vmax = 0.0
        N = 0

        dy = np.diff(y)
        budget = 0.5 * (velocity[1:] + velocity[:-1]) * dy
        prev = np.copy(budget)

        # The following loop uses a kind of bissection approach
        # to look for the suitable value of Vbot.
        while True:

            Vbot = (Vmin + Vmax) / 2.0

            for i, _ in enumerate(y):
                if i > top_idx:
                    velocity[i] = 0.0
                if i >= pt1_idx and i <= top_idx:
                    velocity[i] = Vtop
                if i <= pt2_idx:
                    velocity[i] = Vbot
                if i < pt1_idx and i > pt2_idx:
                    velocity[i] = (Vtop - Vbot) / (y[pt1_idx] - y[pt2_idx]) * (y[i] - y[pt2_idx]) + Vbot

            budget = 0.5 * (velocity[1:] + velocity[:-1]) * dy

            if np.abs(np.sum(budget) - np.sum(prev)) < tol and N > nitmin:
                velocity[top_idx + 1:] = default_vel
                self.budget = np.sum(budget)
                return velocity
            else:
                N += 1
                prev = np.copy(budget)

            if Vtop < 0.0:
                if np.sum(budget) < 0.0:
                    Vmax = Vbot
                else:
                    Vmin = Vbot
            else:
                if np.sum(budget) > 0.0:
                    Vmax = Vbot
                else:
                    Vmin = Vbot

        velocity[top_idx + 1:] = default_vel

        self.budget = np.sum(budget)

        return velocity


def circles_grid(radius, minCoord, maxCoord, npoints=72):
    """ This function creates a set of circles using passive tracers

    Parameters
    ----------

        radius :
            radius of the circles
        minCoord :
            minimum coordinates defining the extent of the grid
        maxCoord :
            maximum coordinates defining the extent of the grid
        npoints :
            number of points used to draw each circle.

    example
    -------

    >>> import UWGeodynamics as GEO
    >>> u = GEO.u

    >>> x_c, y_c = GEO.circles_grid(radius = 2.0 * u.kilometer,
    ...                 minCoord=[Model.minCoord[0], 20. * u.kilometer],
    ...                 maxCoord=[Model.maxCoord[0], 40. * u.kilometer])

    """

    if len(minCoord) == 2:
        # Create points on circle
        angles = np.linspace(0, 360, npoints)
        radius = nd(radius)
        x = radius * np.cos(np.radians(angles))
        y = radius * np.sin(np.radians(angles))

        # Calculate centroids
        xc = np.arange(nd(minCoord[0]), nd(maxCoord[0]) + radius, 2. * radius)
        yc = np.arange(nd(minCoord[1]) + radius, nd(maxCoord[1]), 2. * radius * np.sqrt(3) / 2.)
        xc, yc = np.meshgrid(xc, yc)
        # Shift every other row by radius
        xc[::2, :] = xc[::2, :] + radius

        # Calculate coordinates of all circles points
        points = np.zeros((xc.size, 2))
        points[:, 0] = xc.ravel()
        points[:, 1] = yc.ravel()
        coords = np.zeros((x.size, 2))
        coords[:, 0] = x
        coords[:, 1] = y
        points = points[:, np.newaxis] + coords
        x, y = points[:, :, 0].ravel(), points[:, :, 1].ravel()

        return x, y

    if len(minCoord) == 3:
        # Create points on circle
        theta = np.linspace(0, 180, npoints)
        phi = np.linspace(0, 360, npoints)
        radius = nd(radius)
        theta, phi = np.meshgrid(theta, phi)

        x = radius * np.sin(np.radians(theta.ravel())) * np.cos(np.radians(phi.ravel()))
        y = radius * np.sin(np.radians(theta.ravel())) * np.sin(np.radians(phi.ravel()))
        z = radius * np.cos(np.radians(theta.ravel()))

        # Calculate centroids
        xc = np.arange(nd(minCoord[0]) + radius, nd(maxCoord[0]) + radius, 2. * radius)
        yc = np.arange(nd(minCoord[1]) + radius, nd(maxCoord[1]) + radius, 2. * radius * np.sqrt(3)/2.)
        zc = np.arange(nd(minCoord[2]) + radius, nd(maxCoord[2]) + radius, 2. * radius * np.sqrt(3)/2.)
        xc, yc, zc = np.meshgrid(xc, yc, zc)
        # Shift every other row by radius
        yc[:, ::2, :] += radius
        zc[::2, :, :] += radius

        # Calculate coordinates of all circles points
        points = np.zeros((xc.size, 3))
        points[:, 0] = xc.ravel()
        points[:, 1] = yc.ravel()
        points[:, 2] = zc.ravel()
        coords = np.zeros((x.size, 3))
        coords[:, 0] = x
        coords[:, 1] = y
        coords[:, 2] = y
        points = points[:, np.newaxis] + coords
        x = points[:, :, 0].ravel()
        y = points[:, :, 1].ravel()
        z = points[:, :, 2].ravel()

        return x, y, z


def circle_points_tracers(radius, centre=tuple([0., 0.]), npoints=72):
    angles = np.linspace(0, 360, npoints)
    radius = nd(radius)
    x = radius * np.cos(np.radians(angles)) + nd(centre[0])
    y = radius * np.sin(np.radians(angles)) + nd(centre[1])
    return x, y


def sphere_points_tracers(radius, centre=tuple([0., 0., 0.]), npoints=30):
    theta = np.linspace(0, 180, npoints)
    phi = np.linspace(0, 360, npoints)
    radius = nd(radius)
    theta, phi = np.meshgrid(theta, phi)

    x = radius * np.sin(np.radians(theta.ravel())) * np.cos(np.radians(phi.ravel()))
    y = radius * np.sin(np.radians(theta.ravel())) * np.sin(np.radians(phi.ravel()))
    z = radius * np.cos(np.radians(theta.ravel()))

    x += nd(centre[0])
    y += nd(centre[1])
    z += nd(centre[2])

    return x, y, z


class Nearest_neighbors_projector(object):

    def __init__(self, mesh, swarm, swarm_variable, mesh_variable):
        self.mesh = mesh
        self.swarm = swarm
        self.swarm_variable = swarm_variable
        self.mesh_variable = mesh_variable

    def solve(self):
        tree = spatial.cKDTree(self.swarm.particleCoordinates.data)
        ids = tree.query(self.mesh.data)
        pts = self.swarm.particleCoordinates.data[ids, :]
        self.mesh_variable.data[...] = self.swarm_variable.evaluate(pts)


def fn_Tukey_window(r, centre, width, top, bottom):
    """ Define a tuckey window

    A Tukey window is a rectangular window with the first and last r/2
    percent of the width equal to parts of a cosine.
    see tappered cosine function

    """

    centre = nd(centre)
    width = nd(width)
    top = nd(top)
    bottom = nd(bottom)

    x = fn.input()[0]
    y = fn.input()[1]

    start = centre - 0.5 * width
    xx = (x - start) / width
    x_conditions = [((0. <= xx) & (xx < r /2.0), 0.5 * (1.0 + fn.math.cos(2. * np.pi / r *(xx - r / 2.0)))),
                    ((r / 2.0 <= xx) & (xx < 1.0 - r / 2.0), 1.0),
                    ((1.0 - r / 2.0 <= xx) & (xx <= 1.0), 0.5 * (1. + fn.math.cos(2. * np.pi / r *(xx + r / 2.0)))),
                    (True, 0.0)]

    x_conditions =  fn.branching.conditional(x_conditions)

    y_conditions = fn.branching.conditional([((y >= bottom) & (y <= top), 1.0), (True, 0.0)])
    return x_conditions * y_conditions


class NonLinearBlock(object):
    def __init__(self, string):
        self.string = string
        self.data = dict()
        self.data["Pressure Solve times"] = self.get_vals(["Pressure Solve"], 3)
        self.data["Final V Solve times"] = self.get_vals(["Final V Solve"], 4)
        self.data["Total BSSCR times"] = self.get_vals(["Total BSSCR Linear solve time"], 5)
        self.data["Residuals"] = self.get_vals(["converged", "Residual", "Tolerance"], 5, func=str)
        self.data["Residuals"] = [float(val[:-1]) for val in self.data["Residuals"]]
        self.data["Iterations"] = self.get_vals(["Non linear solver - iteration"], -1, func=int)
        self.data["Solution Time"] = self.get_vals(["solution time"], 5)


    def get_vals(self, FINDSTRING, pos, func=float):
        f = self.string.splitlines()
        vals = [func(line.split()[pos]) for line in f if all([F.lower() in line.lower() for F in FINDSTRING])]
        return vals


class MovingWall(object):

    def __init__(self, velocity):

        self._Model = None
        self._wall = None

        self.velocity = velocity
        self.material = None
        if isinstance(velocity, (list, tuple)):
            self.velocityFn = fn.branching.conditional(velocity)
        else:
            self.velocityFn = fn.misc.constant(nd(velocity))

        self.wall_operators = {"left": op.le,
                          "right": op.ge,
                          "top":op.ge,
                          "bottom": op.le,
                          "front": op.ge,
                          "back": op.le}

        self.wall_direction_axis = {"left": 0,
                               "right": 0,
                               "front": 1,
                               "back": 1,
                               "top": -1,
                               "bottom": -1}

    @property
    def Model(self):
        return self._Model

    @Model.setter
    def Model(self, value):
        self._Model = value
        self.wall_init_pos = {"left": value.minCoord[0],
                         "right": value.maxCoord[0],
                         "front": value.minCoord[1],
                         "back": value.maxCoord[1],
                         "bottom": value.minCoord[-1],
                         "top": value.maxCoord[-1]}

        if value.mesh.dim == 2:
            self.wall_options = {value.left_wall: "left",
                                 value.right_wall: "right",
                                 value.top_wall: "top",
                                 value.bottom_wall: "bottom"}
        else:
            self.wall_options = {value.left_wall: "left",
                                 value.right_wall: "right",
                                 value.front_wall: "front",
                                 value.back_wall: "back",
                                 value.top_wall: "top",
                                 value.bottom_wall: "bottom"}

        # Create a new viscous material with viscosity set to MaxViscosity:
        if not self.material:
            self.material = value.add_material(name="Wall")
            self.material.viscosity = value.maxViscosity
            self.material.density = 0.

    @property
    def wall(self):
        return self._wall

    @wall.setter
    def wall(self, value):
        self._wall = self.wall_options[value]
        self.wallFn = self._create_function()

    def _create_function(self):

        # Create wall function
        operator = self.wall_operators[self._wall]
        axis = self.wall_direction_axis[self._wall]
        pos = self.wall_init_pos[self._wall]
        condition = [(operator(fn.input()[axis],(nd(self.Model.time) *
                                        self.velocityFn +
                                        nd(pos))), True),
                     (True, False)]

        return fn.branching.conditional(condition)

    def get_wall_indices(self):

        # Return new indexSet for the wall
        mesh = self.Model.mesh
        swarm = self.Model.swarm

        nodes = mesh.data_nodegId[self.wallFn.evaluate(mesh)]

        # Update Material Field
        condition = [(self.wallFn, self.material.index), (True, self.Model.materialField)]
        func = fn.branching.conditional(condition)
        self.Model.materialField.data[...] = func.evaluate(swarm)

        axis = self.wall_direction_axis[self.wall]

        return nodes, axis


class LogFile(object):
    def __init__(self, filename):
        self.filename = filename
        self.nonLinear_blocks = self.get_nonLinear_blocks()
        self.pressure_solve_times = list()
        for obj in self.nonLinear_blocks:
            self.pressure_solve_times += obj.data["Pressure Solve times"]
        self.finalV_solve_times = list()
        for obj in self.nonLinear_blocks:
            self.finalV_solve_times += obj.data["Final V Solve times"]
        self.total_BSSCR_times = list()
        for obj in self.nonLinear_blocks:
            self.total_BSSCR_times += obj.data["Total BSSCR times"]
        self.residuals = list()
        for obj in self.nonLinear_blocks:
            self.residuals += obj.data["Residuals"]
        self.iterations = list()
        for obj in self.nonLinear_blocks:
            self.iterations.append(obj.data["Iterations"][-1])
        self.solution_times = list()
        for obj in self.nonLinear_blocks:
            self.solution_times += obj.data["Solution Time"]

    def get_nonLinear_blocks(self):
        non_linear_blocks = list()
        with open(self.filename, "r") as f:
            block = ""
            inBlock = False
            step=0
            for line in f:
                if "Non linear solver" in line:
                    inBlock = True
                    step += 1
                if inBlock:
                    if "Converged" not in line:
                        block += line
                    else:
                        block += line
                        inBlock = False
                        block = NonLinearBlock(block)
                        non_linear_blocks.append(block)
                        block=""
            # Process last potentially non-converged block
            if block:
                block = NonLinearBlock(block)
                non_linear_blocks.append(block)
        self.nonLinear_blocks = non_linear_blocks
        return self.nonLinear_blocks


def extract_profile(field,
                    line,
                    nsamples=1000):
    """ Extract values along a line

    Parameters:
        field: The field to extract the data from
        line: list of (x,y, [z]) coordinates defining the sampling
              line.
        nsamples: number of sampling points
    """

    comm = MPI.COMM_WORLD
    size = comm.Get_size()
    if size > 1:
        raise NotImplementedError("""The extract_profile function will not work
                                  in parallel""")

    coords = np.array([(nd(x), nd(y)) for (x, y) in line])

    x = np.linspace(coords[0, 0], coords[-1, 0], nsamples)

    if coords[0, 0] == coords[-1, 0]:
        y = np.linspace(coords[0, 1], coords[-1, 1], nsamples)
    else:
        from scipy import interpolate
        f = interpolate.interp1d(coords[:, 0], coords[:, 1])
        y = f(x)

    dx = np.diff(x)
    dy = np.diff(y)
    distances = np.zeros(x.shape)
    distances[1:] = np.sqrt(dx**2 + dy**2)
    distances = np.cumsum(distances)

    pts = np.array([x, y]).T
    values = field.evaluate(pts)

    return distances, values



from __future__ import print_function, absolute_import


import json
import abc
import underworld.function as fn
import numpy as np
from .scaling import UnitRegistry as u
from .scaling import nonDimensionalize as nd
from copy import copy
from collections import OrderedDict

ABC = abc.ABCMeta('ABC', (object,), {})


def linearCohesionWeakening(cumulativeTotalStrain, Cohesion, CohesionSw,
                            epsilon1=0.5, epsilon2=1.5):
    """Calculate the cohesion through linear weakening as
       a function of accumulated plastic strain.

    Parameters
    ----------

        cumulativeTotalStrain :
            Accumulated Plastic Strain field
        Cohesion :
            Cohesion for a pristine material Cohesion(epsilon1)
        CohesionSw :
            Cohesion(epsilon2)
        epsilon1 :
            Start of weakening (fraction of accumulated plastic strain)
        epsilon2 : End of weakening (fraction of accumulated plastic strain)

    Returns
    -------
        Underworld function which returns the cohesion as a function of
        accumulated plastic strain.

    """
    cohesionVal = [(cumulativeTotalStrain < epsilon1, Cohesion),
                   (cumulativeTotalStrain > epsilon2, CohesionSw),
                   (True, Cohesion + ((Cohesion - CohesionSw) /
                                      (epsilon1 - epsilon2)) *
                    (cumulativeTotalStrain - epsilon1))]

    return fn.branching.conditional(cohesionVal)


def linearFrictionWeakening(cumulativeTotalStrain, FrictionCoef,
                            FrictionCoefSw, epsilon1=0.5, epsilon2=1.5):
    """Calculate the friction angle through linear weakening as a function of
       accumulated plastic strain.

    Parameters
    ----------

        cumulativeTotalStrain : Underworld Mesh Variable
            Accumulated Plastic Strain field
        friction angle :
            friction angle for a pristine material (friction angle(epsilon1))
        friction angleSw : float
            friction angle(epsilon2)
        epsilon1 : float
            Start of weakening (fraction of accumulated plastic strain)
        epsilon2 : float
            End of weakening (fraction of accumulated plastic strain)

    Returns
    -------

        Underworld function which returns the friction angle as a function of
        accumulated plastic strain.
    """

    frictionVal = [(cumulativeTotalStrain < epsilon1, FrictionCoef),
                   (cumulativeTotalStrain > epsilon2, FrictionCoefSw),
                   (True, FrictionCoef + ((FrictionCoef - FrictionCoefSw) /
                                          (epsilon1 - epsilon2)) *
                    (cumulativeTotalStrain - epsilon1))]

    frictionVal = fn.branching.conditional(frictionVal)

    return fn.math.atan(frictionVal)


class ViscosityLimiter(object):
    """ Viscosity Limiter Class """

    def __init__(self, minViscosity, maxViscosity):

        self.minViscosity = minViscosity
        self.maxViscosity = maxViscosity

    def apply(self, viscosityField):
        """Apply a viscosity limit to a viscosity function.

        Parameters
        ----------
            viscosityField : viscosity function or field

        Returns
        -------
            viscosity function

        """
        if self.maxViscosity and self.minViscosity:
            maxBound = fn.misc.min(viscosityField, nd(self.maxViscosity))
            minMaxBound = fn.misc.max(maxBound, nd(self.minViscosity))
            return minMaxBound
        else:
            return viscosityField


class StressLimiter(object):
    """Stress Limiter Class"""

    def __init__(self, maxStress):
        # Add unit check
        self.maxStress = maxStress

    def apply(self, stress):
        """Apply a stress limit to a stress function.

        Parameters
        ----------
            stress : stress function or field.

        Returns
        -------
            stress function

        """
        maxBound = fn.misc.min(stress, nd(self.maxStress))
        return maxBound


class Rheology(ABC):
    """Rheology Base Class"""

    def __init__(self, pressureField=None, strainRateInvariantField=None,
                 temperatureField=None, viscosityLimiter=None,
                 stressLimiter=None):

        self.pressureField = pressureField
        self.strainRateInvariantField = strainRateInvariantField
        self.temperatureField = temperatureField
        self.stressLimiter = stressLimiter
        self.firstIter = fn.misc.constant(True)

        self.viscosity = None
        self.plasticity = None
        self.cohesion = None
        self.friction = None

        return

    @abc.abstractmethod
    def muEff(self):
        pass


class DruckerPrager(object):
    """The Drucker Prager yield criterion class.

    The yield strength


    """

    def __init__(self, name=None, cohesion=None, frictionCoefficient=None,
                 cohesionAfterSoftening=None,
                 frictionAfterSoftening=None,
                 epsilon1=0.0, epsilon2=0.2):
        """Drucker Prager yield Rheology.

         The pressure-dependent Drucker-Yield criterion is defined as follow:

         .. math::

            :label: druckerprager

            \sigma_y = C \cos\phi + \sin\phi P \quad \text{(2D)}

            \sigma_y = \frac{6C\cos\phi}{\sqrt3(3-\sin\phi)} +
                       \frac{6\sin\phi P}{\sqrt3(3-\sin\phi)} \quad \text{(3D)}


         Setting the friction angle :math:`\phi=0` gives the von Mises Criterion.
         In 2D, equation :eq:`druckerprager` corresponds to the Mohr-Coulomb criterion,
         while in 3D it circumscribes the Mohr-Coulomb yield surface.

         Linear cohesion and friction weakening can be added by defining their
         initial and final values over a range of accumulated plastic strain.

        Parameters
        ----------
            cohesion :
                Cohesion for the pristine material(initial cohesion)
            cohesionAfterSoftening :
                Cohesion of the weakened material
            frictionCoefficient :
                friction angle for a pristine material
            frictionAfterSoftening :
                friction angle of weakened material(epsilon2)
            epsilon1 :
                Start of weakening (fraction of accumulated plastic strain)
            epsilon2 :
                End of weakening (fraction of accumulated plastic strain)

        Returns
        -------

        An UWGeodynamics DruckerPrager class

        """
        self.name = name
        self._cohesion = cohesion
        self._frictionCoefficient = frictionCoefficient
        self.cohesionAfterSoftening = cohesionAfterSoftening
        self.frictionAfterSoftening = frictionAfterSoftening

        self.plasticStrain = None
        self.pressureField = None
        self.epsilon1 = epsilon1
        self.epsilon2 = epsilon2

        self.cohesionWeakeningFn = linearCohesionWeakening
        self.frictionWeakeningFn = linearFrictionWeakening

        self.onlinePDF = None
        self.citation = None

    def __getitem__(self, name):
        return self.__dict__[name]

    def _repr_html_(self):
        attributes = OrderedDict()
        if isinstance(self.citation, str):
            attributes["Citation"] = self.citation
            if isinstance(self.onlinePDF, str):
                if self.onlinePDF.startswith("http"):
                    attributes["Citation"] = '<a href="{0}">{1}</a>'.format(
                        self.onlinePDF, self.citation)
        attributes["Cohesion"] = self.cohesion
        attributes["Cohesion After Softening"] = self.cohesionAfterSoftening
        attributes["Friction Coefficient"] = self.frictionCoefficient
        attributes["Friction Coefficient after Softening"] = (
            self.frictionAfterSoftening)
        attributes["Epsilon 1"] = self.epsilon1
        attributes["Epsilon 2"] = self.epsilon2
        header = "<table>"
        footer = "</table>"
        html = ""
        for key, val in attributes.items():
            html += '<tr><td style="text-align:left;">{0}</td><td style="text-align:left;">{1}</td></tr>'.format(key, val)

        return header + html + footer

    @property
    def cohesion(self):
        return self._cohesion

    @cohesion.setter
    def cohesion(self, value):
        self._cohesion = value
        self._cohesionAfterSoftening = value

    @property
    def cohesionAfterSoftening(self):
        return self._cohesionAfterSoftening

    @cohesionAfterSoftening.setter
    def cohesionAfterSoftening(self, value):
        if value:
            self._cohesionAfterSoftening = value
        else:
            self._cohesionAfterSoftening = self.cohesion

    @property
    def frictionCoefficient(self):
        return self._frictionCoefficient

    @frictionCoefficient.setter
    def frictionCoefficient(self, value):
        self._frictionCoefficient = value
        self._frictionAfterSoftening = value

    @property
    def frictionAfterSoftening(self):
        return self._frictionAfterSoftening

    @frictionAfterSoftening.setter
    def frictionAfterSoftening(self, value):
        if value:
            self._frictionAfterSoftening = value
        else:
            self._frictionAfterSoftening = self.frictionCoefficient

    def _frictionFn(self):
        if self.plasticStrain:
            friction = self.frictionWeakeningFn(
                self.plasticStrain,
                FrictionCoef=self.frictionCoefficient,
                FrictionCoefSw=self.frictionAfterSoftening,
                epsilon1=self.epsilon1,
                epsilon2=self.epsilon2)
        else:
            friction = fn.math.atan(nd(self.frictionCoefficient))
        return friction

    def _cohesionFn(self):
        if self.plasticStrain:
            cohesion = self.cohesionWeakeningFn(
                self.plasticStrain,
                Cohesion=nd(self.cohesion),
                CohesionSw=nd(self.cohesionAfterSoftening))
        else:
            cohesion = fn.misc.constant(self.cohesion)
        return cohesion

    def _get_yieldStress2D(self):
        f = self._frictionFn()
        C = self._cohesionFn()
        P = self.pressureField
        self.yieldStress = (C * fn.math.cos(f) + P * fn.math.sin(f))
        return self.yieldStress

    def _get_yieldStress3D(self):
        f = self._frictionFn()
        C = self._cohesionFn()
        P = self.pressureField
        self.yieldStress = 6.0 * C * fn.math.cos(f)
        self.yieldStress += 6.0 * fn.math.sin(f) * fn.misc.max(P, 0.0)
        self.yieldStress /= (fn.math.sqrt(3.0) * (3.0 + fn.math.sin(f)))
        return self.yieldStress


class VonMises(DruckerPrager):
    """ Von Mises Yield Criterion """

    def __init__(self, name=None, cohesion=None,
                 cohesionAfterSoftening=None,
                 epsilon1=0.0, epsilon2=0.2):

        super(VonMises, self).__init__(cohesion=cohesion,
                                       cohesionAfterSoftening=cohesionAfterSoftening,
                                       frictionCoefficient=0.,
                                       frictionAfterSoftening=0.,
                                       epsilon1=epsilon1, epsilon2=epsilon2)


class ConstantViscosity(Rheology):
    """The newtonian rheology Class."""

    def __init__(self, viscosity):
        """Newtonian Rheology.

        Parameters
        ----------
            Viscosity :
                newtonian viscosity of the material

        Returns
        -------

            An UWGeodynamics ConstantViscosity Class

        """
        super(ConstantViscosity, self).__init__()
        self.viscosity = viscosity
        self.name = "Constant ({0})".format(str(viscosity))

    @property
    def muEff(self):
        return self._effectiveViscosity()

    def _effectiveViscosity(self):
        return fn.Function.convert(nd(self.viscosity))


class ViscousCreep(Rheology):
    """ Viscous Creep Class """

    def __init__(self,
                 name=None,
                 preExponentialFactor=1.0,
                 stressExponent=1.0,
                 activationVolume=0.0,
                 activationEnergy=0.0,
                 waterFugacity=None,
                 grainSize=None,
                 meltFraction=None,
                 grainSizeExponent=0.0,
                 waterFugacityExponent=0.0,
                 meltFractionFactor=0.0,
                 f=1.0,
                 BurgersVectorLength=0.5e-9 * u.metre,
                 mineral="unspecified",
                 creep_type="unspecified"):
        """ Viscous Creep Rheology

        Deformation of materials on long timescale is predominantly achieved
        through viscous diffusion and dislocation creep. Those processes can be
        expressed using the following equation:

        .. math::

           \eta_{\text{eff}}^{vcreep} = \frac{1}{2}A^{\frac{-1}{n}}
           \dot{\epsilon}^{\frac{(1-n)}{n}}d^{\frac{m}{n}}\exp{\left(\frac{E + PV}{nRT}\right)}

        with `A` the prefactor, :math:`\dot{\epsilon}` the square root of the second invariant of the
        deviatoric strain rate tensor, `d` the grain size, `p` the grain size exponent, `E` the
        activation energy, `P` the pressure, `V` the activation volume, `n` the stress exponent,
        `R` the Gas Constant and `T` the temperature.

        Parameters
        ----------

            name : str
                Name of the Viscous Law
            preExponentialFactor :
                pre-exponential factor/ pre-factor
            stressExponent : float
                Stress exponent
            activationVolume :
                Activation Volume
            activationEnergy :
                Activation Energy
            waterFugacity :
                Water Fugacity
            grainSize :
                Grain Size
            meltFraction :
                Initial Fraction of Melt
            grainSizeExponent :
                Grain size eponent
            waterFugacityExponent :
                Water fugacity exponent
            meltFractionFactor :
                Melt Fraction
            f :
                Flow Law scaling factor
            BurgersVectorLength :
                Burger vector length
            mineral :
                Mineral associated to the flow law
            creep_type :
                "Dislocation, Diffusion"

        Returns
        -------

            An UWGeodynamics ViscousCreep Class.

        """

        super(ViscousCreep, self).__init__()

        self.name = name
        self.mineral = mineral
        self.creep_type = creep_type
        self.preExponentialFactor = preExponentialFactor
        self.stressExponent = stressExponent
        self.activationVolume = activationVolume
        self.activationEnergy = activationEnergy
        self.waterFugacity = waterFugacity
        self.grainSize = grainSize
        self.meltFraction = meltFraction
        self.grainSizeExponent = grainSizeExponent
        self.waterFugacityExponent = waterFugacityExponent
        self.meltFractionFactor = meltFractionFactor
        self.f = f
        self.constantGas = 8.3144621 * u.joule / u.mole / u.degK
        self.BurgersVectorLength = BurgersVectorLength

        self.onlinePDF = None
        self.citation = None

    def __mul__(self, other):
        self.f = other
        return self

    def __rmul__(self, other):
        self.f = other
        return self

    def __getitem__(self, name):
        return self.__dict__[name]

    def _repr_html_(self):
        attributes = OrderedDict()
        if isinstance(self.citation, str):
            attributes["Citation"] = self.citation
            if isinstance(self.onlinePDF, str):
                if self.onlinePDF.startswith("http"):
                    attributes["Citation"] = '<a href="{0}">{1}</a>'.format(
                        self.onlinePDF, self.citation)
        attributes["Mineral"] = self.mineral
        attributes["Pre-exponential factor"] = self.preExponentialFactor
        attributes["Stress Exponent"] = self.stressExponent
        attributes["Activation Volume"] = self.activationVolume
        attributes["Activation Energy"] = self.activationEnergy
        attributes["Factor"] = self.f
        attributes["Grain Size"] = self.grainSize
        attributes["Grain Size Exponent"] = self.grainSizeExponent
        attributes["Water Fugacity"] = self.waterFugacity
        attributes["Water Fugacity Exponent"] = self.waterFugacityExponent
        attributes["Melt Fraction"] = self.meltFraction
        attributes["Melt Fraction Factor"] = self.meltFractionFactor
        header = "<table>"
        footer = "</table>"
        html = """
        <tr>
            <th colspan="2" style="text-align:center;">{0}: {1}</th>
        </tr>""".format(self.creep_type, self.name)
        for key, val in attributes.items():
            if val is not None:
                html += '<tr><td style="text-align:left;width:20%;">{0}</td><td style="text-align:left;width:80%">{1}</td></tr>'.format(key, val)


        return header + html + footer

    @property
    def muEff(self):
        return self._effectiveViscosity()

    def _effectiveViscosity(self):
        """
        Return effective viscosity based on

        Power law creep equation
        viscosity = 0.5 * A^(-1/n) * edot_ii^((1-n)/n) * d^(m/n) * exp((E + P*V)/(nRT))

        Parameters:
        -----------

            A:
                prefactor
            I:
                square root of second invariant of deviatoric
                strain rate tensor,
            d:
                grain size,
            p:
                grain size exponent,
            E:
                activation energy,
            P:
                pressure,
            Va:
                activation volume,
            n:
                stress exponent,
            R:
                gas constant,
            T:
                temperature.
        """

        A = nd(self.preExponentialFactor)
        n = nd(self.stressExponent)
        P = self.pressureField
        T = self.temperatureField
        Q = nd(self.activationEnergy)
        Va = nd(self.activationVolume)
        p = nd(self.grainSizeExponent)
        d = nd(self.grainSize)
        r = nd(self.waterFugacityExponent)
        fH2O = nd(self.waterFugacity)
        I = self.strainRateInvariantField
        f = self.f
        #F = self.meltFraction
        #alpha = self.meltFractionFactor
        R = nd(self.constantGas)
        b = nd(self.BurgersVectorLength)

        mu_eff = f * 0.5 * A**(-1.0 / n)

        if np.abs(n - 1.0) > 1e-5:
            mu_eff *= I**((1.0 - n) / n)

        # Grain size dependency
        if p and d:
            mu_eff *= (d / b)**(p / n)

        # Water dependency
        if r and fH2O:
            mu_eff *= fH2O**(-r / n)

        #if F:
        #    mu_eff *= fn.math.exp(-1.0 * alpha * F / n)

        if T:
            mu_eff *= fn.math.exp((Q + P * Va) / (R * T * n))

        return mu_eff


class CompositeViscosity(Rheology):

    def __init__(self, viscosities):

        if not isinstance(viscosities, (list, tuple)):
            raise ValueError('viscosities must be a list of viscosities')

        for viscosity in viscosities:
            if not isinstance(viscosity, ViscousCreep):
                raise ValueError('The viscosity entered is not of ViscousCreep type')

        self.viscosities = viscosities

        self.pressureField = None
        self.strainRateInvariantField = None
        self.temperatureField = None

    @property
    def muEff(self):
        muEff = fn.misc.constant(0.0)
        for viscosity in self.viscosities:
            viscosity.pressureField = self.pressureField
            viscosity.strainRateInvariantField = self.strainRateInvariantField
            viscosity.temperature = self.temperatureField
            muEff += 1.0 / viscosity.muEff

        return 1.0 / muEff


class TemperatureAndDepthDependentViscosity(Rheology):

    def __init__(self, eta0, beta, gamma, reference, temperatureField=None):

        self._eta0 = nd(eta0)
        self._gamma = gamma
        self._beta = gamma
        self._reference = nd(reference)

    @property
    def muEff(self):
        coord = fn.input()
        return (self._eta0 * fn.math.exp(self._gamma *
                                         (coord[-1] - self._reference)))


class ViscousCreepRegistry(object):
    def __init__(self, filename=None):

        if not filename:
            import pkg_resources
            filename = pkg_resources.resource_filename(
                __name__, "ressources/ViscousRheologies.json")

        with open(filename, "r") as infile:
            self._viscousLaws = json.load(infile)

        for key in self._viscousLaws.keys():
            coefficients = self._viscousLaws[key]["coefficients"]
            for key2 in coefficients.keys():
                value = coefficients[key2]["value"]
                units = coefficients[key2]["units"]
                if units != "None":
                    coefficients[key2] = u.Quantity(value, units)
                else:
                    coefficients[key2] = value

        self._dir = {}
        for key in self._viscousLaws.keys():
            mineral = self._viscousLaws[key]["Mineral"]
            rh_type = self._viscousLaws[key]["Type"]
            name = key.replace(",", "").replace(".", "")
            name = [word.strip() for word in name.split()
                    if word.lower() not in ["viscous", "creep"]]
            name = "_".join(name)
            self._dir[name] = ViscousCreep(
                name=key, mineral=mineral, creep_type=rh_type,
                **self._viscousLaws[key]["coefficients"])

            try:
                self._dir[name].onlinePDF = self._viscousLaws[key]["onlinePDF"]
            except KeyError:
                pass

            try:
                self._dir[name].citation = self._viscousLaws[key]["citation"]
            except KeyError:
                pass

    def __dir__(self):
        # Make all the rheology available through autocompletion
        return list(self._dir.keys())

    def __getattr__(self, item):
        # Make sure to return a new instance of ViscousCreep
        return copy(self._dir[item])


class PlasticityRegistry(object):
    def __init__(self, filename=None):

        if not filename:
            import pkg_resources
            filename = pkg_resources.resource_filename(
                __name__, "ressources/PlasticRheologies.json")

        with open(filename, "r") as infile:
            _plasticLaws = json.load(infile)

        for key in _plasticLaws.keys():
            coefficients = _plasticLaws[key]["coefficients"]
            for key2 in coefficients.keys():
                value = coefficients[key2]["value"]
                units = coefficients[key2]["units"]
                if units != "None":
                    coefficients[key2] = u.Quantity(value, units)
                else:
                    coefficients[key2] = value

        self._dir = {}
        for key in _plasticLaws.keys():
            name = key.replace(" ", "_").replace(",", "").replace(".", "")
            name = name.replace(")", "").replace("(", "")
            self._dir[name] = DruckerPrager(
                name=key, **_plasticLaws[key]["coefficients"])

            try:
                self._dir[name].onlinePDF = _plasticLaws[key]["onlinePDF"]
            except KeyError:
                pass

            try:
                self._dir[name].citation = _plasticLaws[key]["citation"]
            except KeyError:
                pass

    def __dir__(self):
        # Make all the rheology available through autocompletion
        return list(self._dir.keys())

    def __getattr__(self, item):
        # Make sure to return a new instance of ViscousCreep
        return copy(self._dir[item])


class Elasticity(Rheology):

    def __init__(self, shear_modulus, observation_time):
        super(Elasticity, self).__init__()
        self.shear_modulus = shear_modulus
        self.observation_time = observation_time
        self.previousStress = None

    @property
    def muEff(self):
        return self._effectiveViscosity()

    def _effectiveViscosity(self):
        if not self.viscosity:
            raise ValueError("Can not find viscosity field")

        # Maxwell relaxation time
        alpha = self.viscosity / nd(self.shear_modulus)
        # observation time
        dt_e = nd(self.observation_time)
        # Calculate effective viscosity
        mu_eff = (self.viscosity * dt_e) / (alpha + dt_e)
        return mu_eff

    @property
    def elastic_stress(self):
        return self._elastic_stress()

    def _elastic_stress(self):
        # Check that the viscosity field has been properly
        # linked
        if not self.viscosity:
            raise ValueError("Can not find viscosity field")
        if not self.previousStress:
            raise ValueError("Can not find previous stress field")

        elasticStressFn = self.viscosity / (nd(self.shear_modulus) *
                                            nd(self.observation_time))
        elasticStressFn *= self.previousStress
        return elasticStressFn



from __future__ import print_function,  absolute_import
from .scaling import nonDimensionalize as nd
from .scaling import UnitRegistry as u
from .scaling import Dimensionalize
import json
from copy import copy


class _Polynom(object):

    def __init__(self, A1, A2=0., A3=0., A4=0.):

        self.A1 = A1
        self.A2 = A2
        self.A3 = A3
        self.A4 = A4

    def temperature(self, pressure):
        T = nd(self.A1)
        T += nd(self.A2) * pressure
        T += nd(self.A3) * pressure**2
        T += nd(self.A4) * pressure**3
        return T

    def plot(self, pressure):
        import pylab as plt
        temperature = Dimensionalize(self.temperature(pressure), u.kelvin)
        pressure = Dimensionalize(pressure, u.pascal)
        plt.plot(temperature, pressure)
        plt.gca().invert_yaxis()
        plt.show()


class Solidus(_Polynom):
    """ This class defines a solidus using the
    form suggested by Hirshmann, 2000"""

    def __init__(self, A1, A2=0., A3=0., A4=0.):
        super(Solidus, self).__init__(A1, A2, A3, A4)


class Liquidus(_Polynom):
    """ This class defines a liquidus using the
    form suggested by Hirshmann, 2000"""

    def __init__(self, A1, A2=0., A3=0., A4=0.):
        super(Liquidus, self).__init__(A1, A2, A3, A4)


class SolidusRegistry(object):
    """SolidusRegistry Class"""
    def __init__(self, filename=None):
        """Create a regustry of Solidus polynomials

        Parameters
        ----------

        filename : Name of the json file database

        Returns
        -------
        """

        if not filename:
            import pkg_resources
            filename = pkg_resources.resource_filename(
                __name__, "ressources/Solidus.json")

        with open(filename, "r") as infile:
            _solidii = json.load(infile)

        for key in _solidii.keys():
            coefficients = _solidii[key]["coefficients"]
            for key2 in coefficients.keys():
                value = coefficients[key2]["value"]
                units = coefficients[key2]["units"]
                if units != "None":
                    coefficients[key2] = u.Quantity(value, units)
                else:
                    coefficients[key2] = value

        self._dir = {}
        for key in _solidii.keys():
            name = key.replace(" ", "_").replace(",", "").replace(".", "")
            name = name.replace(")", "").replace("(", "")
            self._dir[name] = Solidus(**_solidii[key]["coefficients"])

    def __dir__(self):
        # Make all the rheology available through autocompletion
        return list(self._dir.keys())

    def __getattr__(self, item):
        # Make sure to return a new instance of ViscousCreep
        return copy(self._dir[item])


class LiquidusRegistry(object):
    """LiquidusRegistry class"""
    def __init__(self, filename=None):
        """Create a regustry of Liquidus polynomials

        Parameters
        ----------

        filename : Name of the json file database

        Returns
        -------
        """

        if not filename:
            import pkg_resources
            filename = pkg_resources.resource_filename(
                __name__, "ressources/Liquidus.json")

        with open(filename, "r") as infile:
            _liquidii = json.load(infile)

        for key in _liquidii.keys():
            coefficients = _liquidii[key]["coefficients"]
            for key2 in coefficients.keys():
                value = coefficients[key2]["value"]
                units = coefficients[key2]["units"]
                if units != "None":
                    coefficients[key2] = u.Quantity(value, units)
                else:
                    coefficients[key2] = value

        self._dir = {}
        for key in _liquidii.keys():
            name = key.replace(" ", "_").replace(",", "").replace(".", "")
            name = name.replace(")", "").replace("(", "")
            self._dir[name] = Liquidus(**_liquidii[key]["coefficients"])

    def __dir__(self):
        # Make all the rheology available through autocompletion
        return list(self._dir.keys())

    def __getattr__(self, item):
        # Make sure to return a new instance of ViscousCreep
        return copy(self._dir[item])



from __future__ import print_function,  absolute_import
from pint import UnitRegistry
from itertools import chain
from collections import OrderedDict
import functools
from pint.compat import zip_longest
from pint.errors import DimensionalityError

try:              # Python 2
    str_base = basestring
    items = 'iteritems'
except NameError: # Python 3
    str_base = str, bytes, bytearray
    items = 'items'

_RaiseKeyError = object()  # singleton for no-default behavior

u = UnitRegistry = UnitRegistry()


def ensure_lower(maybe_str):
    """dict keys can be any hashable object - only call lower if str"""
    return maybe_str.lower() if isinstance(maybe_str, str_base) else maybe_str


def ensure_to_base_units(val):
    return val.to_base_units()


class TransformedDict(dict):  # dicts take a mapping or iterable as their optional first argument
    __slots__ = () # no __dict__ - that would be redundant

    @staticmethod # because this doesn't make sense as a global function.
    def _process_args(mapping=(), **kwargs):
        if hasattr(mapping, items):
            mapping = getattr(mapping, items)()
        return ((ensure_lower(k), ensure_to_base_units(v)) for k, v in chain(mapping, getattr(kwargs, items)()))

    def __init__(self, mapping=(), **kwargs):
        super(TransformedDict, self).__init__(self._process_args(mapping, **kwargs))

    def __getitem__(self, k):
        return super(TransformedDict, self).__getitem__(ensure_lower(k))

    def __setitem__(self, k, v):
        return super(TransformedDict, self).__setitem__(ensure_lower(k), ensure_to_base_units(v))

    def __delitem__(self, k):
        return super(TransformedDict, self).__delitem__(ensure_lower(k))

    def get(self, k, default=None):
        return super(TransformedDict, self).get(ensure_lower(k), default)

    def setdefault(self, k, default=None):
        return super(TransformedDict, self).setdefault(ensure_lower(k), default)

    def pop(self, k, v=_RaiseKeyError):
        if v is _RaiseKeyError:
            return super(TransformedDict, self).pop(ensure_lower(k))
        return super(TransformedDict, self).pop(ensure_lower(k), v)

    def update(self, mapping=(), **kwargs):
        super(TransformedDict, self).update(self._process_args(mapping, **kwargs))

    def __contains__(self, k):
        return super(TransformedDict, self).__contains__(ensure_lower(k))

    def copy(self): # don't delegate w/ super - dict.copy() -> dict :(
        return type(self)(self)

    @classmethod
    def fromkeys(cls, keys, v=None):
        return super(TransformedDict, cls).fromkeys((ensure_lower(k) for k in keys), v)

    def _repr_html_(self):
        attributes  = OrderedDict()
        attributes["[mass]"] = self["[mass]"]
        attributes["[length]"] = self["[length]"]
        attributes["[temperature]"] = self["[temperature]"]
        attributes["[time]"] = self["[time]"]
        attributes["[substance]"] = self["[substance]"]
        header = "<table>"
        footer = "</table>"
        html = ""
        for key, val in attributes.items():
            html += "<tr><td>{0}</td><td>{1}</td></tr>".format(key, val)

        return header + html + footer



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
import os
import subprocess
from os import path

MAJOR               = 2
MINOR               = 6
MICRO               = 1
ISRELEASED          = True
VERSION             = '%d.%d.%d' % (MAJOR, MINOR, MICRO)

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


# Write the default uwgeodynamics file
with open('uwgeodynamicsrc.template') as fd:
    template = fd.read()
with open('UWGeodynamics/uwgeo-data/uwgeodynamicsrc', 'w') as fd:
    fd.write(template)


# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout = subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "Unknown"

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse','--abbrev-ref', 'HEAD'])
        GIT_BRANCH = out.strip().decode('ascii')
    except OSError:
        GIT_BRANCH = "Unknown"

    return GIT_REVISION, GIT_BRANCH

def get_version_info():
    # Adding the git rev number needs to be done inside write_version_py(),
    # otherwise the import of numpy.version messes up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists('.git'):
        GIT_REVISION, GIT_BRANCH = git_version()
    elif os.path.exists('UWGeodynamics/version.py'):
        # must be a source distribution, use existing version file
        try:
            from UWGeodynamics.version import git_revision as GIT_REVISION
        except ImportError:
            raise ImportError("Unable to import git_revision. Try removing " \
                              "UWGeodynamics/version.py and the build directory " \
                              "before building.")
        try:
            from UWGeodynamics.version import git_branch as GIT_BRANCH
        except ImportError:
            raise ImportError("Unable to import git_branch. Try removing " \
                              "UWGeodynamics/version.py and the build directory " \
                              "before building.")
    else:
        GIT_REVISION = "Unknown"
        GIT_BRANCH = "Unknown"

    if not ISRELEASED:
        FULLVERSION += "-dev-" + GIT_REVISION[:7] + "(" + GIT_BRANCH + ")"

    return FULLVERSION, GIT_REVISION


def write_version_py(filename='UWGeodynamics/version.py'):
    cnt = """
# THIS FILE IS GENERATED FROM UWGeodynamics SETUP.PY
#
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_revision = '%(git_revision)s'
release = %(isrelease)s

if not release:
    version = full_version
"""
    FULLVERSION, GIT_REVISION = get_version_info()

    try:
        a = open(filename, 'w')
        a.write(cnt % {'version': VERSION,
                       'full_version': FULLVERSION,
                       'git_revision': GIT_REVISION,
                       'isrelease': str(ISRELEASED)})
    finally:
        a.close()

write_version_py()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='UWGeodynamics',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=VERSION,  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='Underworld Geodynamics wrapper',  # Required

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://github.com/rbeucher/UWGeodynamics.git',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='Romain Beucher',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='romain.beucher@unimelb.edu.au',  # Optional

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='geodynamics Underworld geophysics geology',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['pint', 'numpy', 'scipy', 'pyshp'],  # Optional

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    #extras_require={  # Optional
    #    'dev': ['check-manifest'],
    #    'test': ['coverage'],
    #},

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    package_data={
        'UWGeodynamics': ['ressources/*', 'uwgeo-data/*'],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    #
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('examples', ['./examples/*']),
    #            ('tutorials', ['./tutorials/*']),
    #            ('manual', ['./manual/*'])],  # Optional

    include_package_data=True,
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    entry_points={  # Optional
        'console_scripts': [
            'UW_to_ASCII=UWGeodynamics.utilities:main',
        ],
    },
)



from __future__ import print_function,  absolute_import
from .scaling import u
from pint import UndefinedUnitError
import six
try:
    import collections.abc as abc
except ImportError:
    # python 2
    import collections as abc


def validate_gravity(s):
    return u.check('[length]')(s)


def _listify_validator(scalar_validator, allow_stringlist=False):
    def f(s):
        if isinstance(s, six.string_types):
            try:
                return [scalar_validator(v.strip()) for v in s.split(',')
                        if v.strip()]
            except Exception:
                if allow_stringlist:
                    # Sometimes, a list of colors might be a single string
                    # of single-letter colornames. So give that a shot.
                    return [scalar_validator(v.strip()) for v in s if v.strip()]
                else:
                    raise
        # We should allow any generic sequence type, including generators,
        # Numpy ndarrays, and pandas data structures.  However, unordered
        # sequences, such as sets, should be allowed but discouraged unless the
        # user desires pseudorandom behavior.
        elif isinstance(s, abc.Iterable) and not isinstance(s, abc.Mapping):
            # The condition on this list comprehension will preserve the
            # behavior of filtering out any empty strings (behavior was
            # from the original validate_stringlist()), while allowing
            # any non-string/text scalar values such as numbers and arrays.
            return [scalar_validator(v) for v in s
                    if not isinstance(v, six.string_types) or v]
        else:
            msg = "{0!r} must be of type: string or non-dictionary iterable.".format(s)
            raise ValueError(msg)
    f.__doc__ = scalar_validator.__doc__
    return f


def validate_quantity(s):
    # Convert to quantity
    s = u.Quantity(s)
    if s.dimensionless:
        return s.magnitude
    return s


def validate_float(s):
    try:
        return float(s)
    except ValueError:
        raise ValueError("Could not convert value to float")


def validate_solver(s):
    if s in ["mg", "mumps", "lu", "superlu", "superludist"]:
        return s
    else:
        raise ValueError("Wrong solver option")


def validate_int(s):
    try:
        return int(s)
    except ValueError:
        raise ValueError("Could not convert value to int")


def validate_int_or_none(s):
    if s:
        return validate_int(s)
    if s is None:
        return
    else:
        raise ValueError("Must be int or None")


def validate_path(s):
    return str(s)


def validate_bool(b):
    """Convert b to a boolean or raise"""
    if isinstance(b, six.string_types):
        b = b.lower()
    if b in ('t', 'y', 'yes', 'on', 'true', '1', 1, True):
        return True
    elif b in ('f', 'n', 'no', 'off', 'false', '0', 0, False):
        return False
    else:
        raise ValueError('Could not convert "%s" to boolean' % b)


def validate_string(s):
    return s


def validate_any(s):
    return s


def validate_viscosity(s):
    try:
        s = u.Quantity(s)
    except UndefinedUnitError:
        try:
            from UWGeodynamics import rheologies
            s = s.replace(" ", "_")
            s = s.replace(",", "")
            s = s.replace(".", "")

            if s in rheologies._dir:
                return rheologies._dir[s]
        except KeyError:
            raise ValueError(
                """Can not find {0} rheology in databases""".format(s))

def validate_averaging(s):
    options = {"arithmetic": 1,
               "geometric": 0,
               "harmonic": -1,
               "maximum": 30,
               "minimum": -30,
               "root mean square": 2}
    if s not in options.keys():
        raise ValueError(
            """{0} is not a valid option, valid options are {1}""".format(
                s, options.keys()))

    return options[s]

validate_stringlist = _listify_validator(six.text_type)
validate_stringlist.__doc__ = 'return a list'
